package ch.ethz.karto.map3d;

import com.sun.opengl.util.texture.TextureData;
import java.awt.image.BufferedImage;
import java.nio.Buffer;
import javax.media.opengl.GL;

/**
 * A 1D or 2D texture
 * @author Bernhard Jenny, Institute of Cartography, ETH Zurich.
 */
public class Map3DTexture {

    /**
     * The cashed texture image. JOGL may dispose the old context and pass a new
     * context to GLEventListener.init(). In this case, the texture must be
     * reloaded by GLEventListener.init(). Reloading can be triggerd by calling
     * disableTexture(). This image can be null.
     */
    private BufferedImage image = null;
    /**
     * The OpenGL name of the texture. 0 means undefined, other values indicate
     * a texture loaded to OpenGL.
     */
    private int[] name = new int[1];
    /**
     * Whether or not this texture is applied.
     */
    private boolean enabled = true;
    /**
     * If true, the name has to be released from OpenGL. The image will be
     * retained in this object and can be reloaded.
     */
    private boolean willUnload = false;

    private GL gl_; // FIXME

    /**
     * Returns an estimation of the maximum size of a texture.
     * @return The maximum dimension of a texture or -1 if an error occurs.
     */
    public static int getMaxTextureSize() {
        return Map3DGLCapabilities.getMaxTextureSize();
    }

    public Map3DTexture() {
    }

    /**
     * Returns true if the texture is two-dimensional, i.e. the texture image
     * consists of more than one row.
     * @return true if this is a two-dimensional texture.
     */
    public boolean is2D() {
        return image != null && image.getHeight() > 1;
    }

    /**
     * Returns true if the texture is one-dimensional, i.e. the texture image
     * consists of only a single row.
     * @return true if this is a one-dimensional texture.
     */
    public boolean is1D() {
        return !is2D();
    }

    /**
     * Returns the OpenGL texture type.
     * @return GL.GL_TEXTURE_1D or GL.GL_TEXTURE_2D
     */
    public int getTexType() {
        return is1D() ? GL.GL_TEXTURE_1D : GL.GL_TEXTURE_2D;
    }

    /**
     * Returns the dimension of the texture.
     * @return 1 or 2
     */
    public int getDim() {
        return is1D() ? 1 : 2;
    }

    /**
     * Enables or disables the use of textures for old fixed function pipeline.
     * This is not needed for shader programs.
     * If both 2D and a 1D texturing are enable, OpenGL uses the 2D texture and
     * ignores the 1D texture.
     * @param gl
     */
    public void updateEnabledState(GL gl) {
        if (isLoaded() && enabled) {
            if (is1D()) {
                gl.glDisable(GL.GL_TEXTURE_2D);
                gl.glEnable(GL.GL_TEXTURE_1D);
            } else {
                gl.glDisable(GL.GL_TEXTURE_1D);
                gl.glEnable(GL.GL_TEXTURE_2D);
            }
        } else {
            gl.glDisable(GL.GL_TEXTURE_1D);
            gl.glDisable(GL.GL_TEXTURE_2D);
        }
    }

    /**
     * Reload the texture image to OpenGL. This is required after the OpenGL
     * context becomes invalid.
     * @param gl
     * @return True if the texture is loaded as a result of calling this method.
     */
    public boolean reload(GL gl) {
        this.willUnload = true;
        return this.loadIfRequired(gl);
    }

    /**
     * Loads the texture to OpenGL if necessary.
     * @param gl
     * @return True if the texture is loaded as a result of calling this method.
     */
    public boolean loadIfRequired(GL gl) {

        // test whether the current textureID needs to be released,
        // which is the case after a call to disableTexture().
        // glDeleteTextures silently ignores 0's and names that do not
        // correspond to existing textures
        if (this.willUnload) {
            if (gl_ == gl) {
                gl.glDeleteTextures(1, name, 0);
            }
            this.name[0] = 0;
            this.willUnload = false;
        }

        if (enabled && image != null && !isLoaded()) {
            // generate a new texture name
            gl.glGenTextures(1, name, 0);
            this.gl_ = gl;

            // bind and load the texture
            this.bindTexture(gl);

            gl.glTexParameteri(getTexType(), GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR);
            gl.glTexParameteri(getTexType(), GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR);

            // GL_CLAMP_TO_EDGE is available only if the GL version is 1.2 or greater.
            if (Map3DGLCapabilities.hasOpenGLVersion(1, 2)) {
                gl.glTexParameteri(getTexType(), GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE);
                gl.glTexParameteri(getTexType(), GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_EDGE);
            }

            TextureData texData = new TextureData(0, 0, false, image);

            int w = texData.getWidth();
            int h = texData.getHeight();
            int f = texData.getPixelFormat();
            int t = texData.getPixelType();
            Buffer b = texData.getBuffer();
            int imgFormat = texData.getInternalFormat();
            if (is2D()) {
                gl.glTexImage2D(GL.GL_TEXTURE_2D, 0, imgFormat, w, h, 0, f, t, b);
            } else {
                gl.glTexImage1D(GL.GL_TEXTURE_1D, 0, imgFormat, w, 0, f, t, b);
            }

            return true;
        }

        return false;
    }

    /**
     * Returns whether a texture has been loaded to OpenGL
     * @return 
     */
    public boolean isLoaded() {
        return name[0] != 0;
    }

    /**
     * Set the texture image. Pass null to release the current image.
     * This also implicetly call setEnabled(true).
     * @param image The new texture image.
     */
    public void setImage(BufferedImage image) {
        this.image = image;
        this.enabled = image != null;
        this.willUnload = true;
    }

    /**
     * Removes the current texture from OpenGL. This sets the willUnload
     * flag. The texture is removed the next time loadIfRequired() is called. 
     * This will not release the cashed image from this object.
     */
    public void disableTexture() {
        this.willUnload = true;
        this.enabled = false;
    }
    
    public void bindTexture(GL gl) {
        if (name[0] != 0 && gl_ == gl) {
            gl.glBindTexture(getTexType(), name[0]);
        }
    }
    
    /**
     * Whether or not the texture is applied.
     * @return the enabled
     */
    public boolean isEnabled() {
        return enabled;
    }

    /**
     * Whether or not the texture is applied.
     * @param enabled the enabled to set
     */
    public void setEnabled(boolean enabled) {
        this.enabled = enabled;
    }

    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        sb.append("Texture ");
        sb.append(isLoaded() ? "loaded" : "not loaded");
        sb.append(is2D() ? "2D" : "1D");
        return sb.toString();
    }

}
