
package ch.ethz.karto.map3d;

import ika.geo.GeoGrid;
import ika.geo.grid.GridInvalidateOperator;
import ika.geo.grid.GridToImageOperator;
import java.awt.image.BufferedImage;
import java.nio.FloatBuffer;
import java.util.Arrays;
import javax.media.opengl.GL;
import javax.media.opengl.GLException;
import javax.media.opengl.glu.GLU;

/**
 * Utility for reading the z-buffer. Based on com.sun.opengl.util.Screenshot.java
 * @author jenny
 */
public class OpenGLDepthBuffer {

    private static int glGetInteger(GL gl, int pname, int[] tmp) {
        gl.glGetIntegerv(pname, tmp, 0);
        return tmp[0];
    }

    /**
     * Returns buffer of floating point values with the lowest row first.
     * @param width
     * @param height
     * @return
     * @throws GLException
     */
    public static FloatBuffer readToBuffer(int width, int height) throws GLException {
        
        GL gl = GLU.getCurrentGL();

        // Set up pixel storage modes
        PixelStorageModes psm = new PixelStorageModes();
        psm.save(gl);

        FloatBuffer fbuffer = FloatBuffer.allocate(width * height);
        gl.glReadPixels(0, 0, width, height,
                GL.GL_DEPTH_COMPONENT,
                GL.GL_FLOAT,
                fbuffer);

        // Restore pixel storage modes
        psm.restore(gl);

        return fbuffer;
    }

    /**
     * Returns a two-dimensional array containing the z-buffer with the
     * highest row first.
     * @param width
     * @param height
     * @return
     * @throws GLException
     */
    public static float[][] readToGrid(int width, int height) throws GLException {
        FloatBuffer fbuffer = readToBuffer(width, height);
        float[] parr = fbuffer.array();
        
        // copy 1D to 2D and flip array vertically for correct results
        float[][] grid = new float[height][width];
        for (int r = 0; r < height; r++) {
            grid[r] = Arrays.copyOfRange(parr, (height - 1 - r) * width, (height - r) * width);
        }
        return grid;
    }

    /**
     * Returns a gray scale image containing the z-buffer with the
     * highest row first.
     * @param width
     * @param height
     * @return
     * @throws GLException
     */
    public static BufferedImage readToImage(int width, int height,
            boolean bit16) throws GLException {

        GeoGrid grid = new GeoGrid(readToGrid(width, height), 1);
        grid = new GridInvalidateOperator(1).operate(grid);
        return new GridToImageOperator(bit16).operate(grid).getBufferedImage();

    }

    static class PixelStorageModes {

        int packAlignment;
        int packRowLength;
        int packSkipRows;
        int packSkipPixels;
        int packSwapBytes;
        int[] tmp = new int[1];

        void save(GL gl) {
            packAlignment = glGetInteger(gl, GL.GL_PACK_ALIGNMENT, tmp);
            packRowLength = glGetInteger(gl, GL.GL_PACK_ROW_LENGTH, tmp);
            packSkipRows = glGetInteger(gl, GL.GL_PACK_SKIP_ROWS, tmp);
            packSkipPixels = glGetInteger(gl, GL.GL_PACK_SKIP_PIXELS, tmp);
            packSwapBytes = glGetInteger(gl, GL.GL_PACK_SWAP_BYTES, tmp);

            gl.glPixelStorei(GL.GL_PACK_ALIGNMENT, 1);
            gl.glPixelStorei(GL.GL_PACK_ROW_LENGTH, 0);
            gl.glPixelStorei(GL.GL_PACK_SKIP_ROWS, 0);
            gl.glPixelStorei(GL.GL_PACK_SKIP_PIXELS, 0);
            gl.glPixelStorei(GL.GL_PACK_SWAP_BYTES, 0);
        }

        void restore(GL gl) {
            gl.glPixelStorei(GL.GL_PACK_ALIGNMENT, packAlignment);
            gl.glPixelStorei(GL.GL_PACK_ROW_LENGTH, packRowLength);
            gl.glPixelStorei(GL.GL_PACK_SKIP_ROWS, packSkipRows);
            gl.glPixelStorei(GL.GL_PACK_SKIP_PIXELS, packSkipPixels);
            gl.glPixelStorei(GL.GL_PACK_SWAP_BYTES, packSwapBytes);
        }
    }
}
