/*
 * GridToImageOperator.java
 *
 * Created on February 3, 2006, 11:40 AM
 *
 */
package ika.geo.grid;

import ika.geo.*;
import java.awt.image.*;
import java.util.Arrays;

/**
 *
 * @author Bernhard Jenny, Institute of Cartography, ETH Zurich.
 */
public class GridToImageOperator implements GridOperator {

    private boolean bit16 = false;

    /** Creates a new instance of GridToImageOperator */
    public GridToImageOperator() {
    }

    public GridToImageOperator(boolean bit16) {
        this.bit16 = bit16;
    }

    @Override
    public String getName() {
        return "Grid To Image";
    }

    @Override
    public GeoImage operate(GeoGrid geoGrid) {

        if (geoGrid == null) {
            throw new IllegalArgumentException();
        }

        final int nrows = geoGrid.getRows();
        final int ncols = geoGrid.getCols();
        if (nrows == 0 || ncols == 0) {
            return null;
        }

        final float[] minMax = geoGrid.getMinMax();
        final float min = minMax[0];
        final float dataRange = minMax[1] - minMax[0];

        float[][] srcGrid = geoGrid.getGrid();

        // Create a BufferedIamge of the gray values in bytes.
        BufferedImage bufferedImage = new BufferedImage(ncols, nrows,
                bit16 ? BufferedImage.TYPE_USHORT_GRAY : BufferedImage.TYPE_BYTE_GRAY);

        // Get the writable raster so that data can be changed.
        WritableRaster wr = bufferedImage.getRaster();

        if (!bit16 || dataRange == 0.f) { // write 8 bit data
            byte[] bytesPixel = new byte[nrows * ncols];
            int id = 0;
            if (dataRange != 0) {
                for (int row = 0; row < nrows; ++row) {
                    float[] srcRow = srcGrid[row];
                    for (int col = 0; col < ncols; ++col) {
                        bytesPixel[id++] = (byte) ((srcRow[col] - min) / dataRange * 255.f);
                    }
                }
            } else {
                // create white image
                Arrays.fill(bytesPixel, (byte) 255);
            }

            // write the byte data to the raster
            wr.setDataElements(0, 0, ncols, nrows, bytesPixel);
        } else {    // write 16 bit data
            short[] shortPixel = new short[nrows * ncols];
            int id = 0;
            if (dataRange != 0) {
                for (int row = 0; row < nrows; ++row) {
                    float[] srcRow = srcGrid[row];
                    for (int col = 0; col < ncols; ++col) {
                        shortPixel[id++] = (short) ((srcRow[col] - min) / dataRange * 65535.f);
                    }
                }
            }

            // write the byte data to the raster
            wr.setDataElements(0, 0, ncols, nrows, shortPixel);
        }

        final double cellSize = geoGrid.getCellSize();
        final double gridWest = geoGrid.getWest() - cellSize / 2.;
        final double gridNorth = geoGrid.getNorth() - cellSize / 2.;
        return new GeoImage(bufferedImage, gridWest, gridNorth, cellSize);
    }
}
